from setuptools import setup

setup(name              = "llrflibs",
      version           = "1.0.1",
      maintainer        = "Zheqiao Geng",
      maintainer_email  = "zheqiao.geng@psi.ch",
      author            = "Zheqiao Geng",
      author_email      = "zheqiao.geng@psi.ch",
      description       = "LLRF algorithm libraries in Python.",
      url               = "https://github.com/aaqiao/LLRFLibsPy",
      license           = "BSD 3-Clause",
      packages          = ['llrflibs'],
      zip_safe          = False)

